@@ -1,7 +1,7 @@
 /*
    PCRE-OCAML - Perl Compatibility Regular Expressions for OCaml
 
-   Copyright (C) 1999-2006  Markus Mottl
+   Copyright (C) 1999-  Markus Mottl
    email: markus.mottl@gmail.com
    WWW:   http://www.ocaml.info
 
@@ -20,14 +20,26 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-#if defined(_WIN32) && defined(_DLL)
-#  define PCREextern __declspec(dllexport)
+#if defined(_WIN32)
+#  define snprintf _snprintf
+#  if defined(_DLL)
+#    define PCREextern __declspec(dllexport)
+#  else
+#    define PCREextern
+#  endif
+#endif
+
+#if __GNUC__ >= 3
+# define inline inline __attribute__ ((always_inline))
+# define __unused __attribute__ ((unused))
 #else
-#  define PCREextern
+# define __unused
+# define inline
 #endif
 
 #include <ctype.h>
 #include <string.h>
+#include <stdio.h>
 
 #include <caml/mlvalues.h>
 #include <caml/alloc.h>
@@ -37,6 +49,17 @@
 
 #include <pcre.h>
 
+/* Error codes as defined for pcre 7.9, undefined in pcre 4.5 */
+#ifndef PCRE_ERROR_PARTIAL
+#define PCRE_ERROR_PARTIAL        (-12)
+#endif
+#ifndef PCRE_ERROR_BADPARTIAL
+#define PCRE_ERROR_BADPARTIAL     (-13)
+#endif
+#ifndef PCRE_ERROR_RECURSIONLIMIT
+#define PCRE_ERROR_RECURSIONLIMIT (-21)
+#endif
+
 typedef const unsigned char *chartables;  /* Type of chartable sets */
 
 /* Contents of callout data */
@@ -47,14 +70,7 @@ struct cod {
 };
 
 /* Cache for exceptions */
-static value *pcre_exc_Not_found     = NULL;  /* Exception [Not_found] */
-static value *pcre_exc_Partial       = NULL;  /* Exception [Partial] */
-static value *pcre_exc_BadPartial    = NULL;  /* Exception [BadPartial] */
-static value *pcre_exc_BadPattern    = NULL;  /* Exception [BadPattern] */
-static value *pcre_exc_BadUTF8       = NULL;  /* Exception [BadUTF8] */
-static value *pcre_exc_BadUTF8Offset = NULL;  /* Exception [BadUTF8Offset] */
-static value *pcre_exc_InternalError = NULL;  /* Exception [InternalError] */
-static value *pcre_exc_MatchLimit    = NULL;  /* Exception [MatchLimit] */
+static value *pcre_exc_Error         = NULL;  /* Exception [Error] */
 static value *pcre_exc_Backtrack     = NULL;  /* Exception [Backtrack] */
 
 /* Cache for polymorphic variants */
@@ -120,23 +136,17 @@ static int pcre_callout_handler(pcre_callout_block* cb)
 
 /* Fetchs the named OCaml-values + caches them and
    calculates + caches the variant hash values */
-CAMLprim value pcre_ocaml_init(value unit)
+CAMLprim value pcre_ocaml_init(value __unused v_unit)
 {
-  pcre_exc_Not_found     = caml_named_value("Pcre.Not_found");
-  pcre_exc_Partial       = caml_named_value("Pcre.Partial");
-  pcre_exc_BadPartial    = caml_named_value("Pcre.BadPartial");
-  pcre_exc_BadPattern    = caml_named_value("Pcre.BadPattern");
-  pcre_exc_BadUTF8       = caml_named_value("Pcre.BadUTF8");
-  pcre_exc_InternalError = caml_named_value("Pcre.InternalError");
-  pcre_exc_MatchLimit    = caml_named_value("Pcre.MatchLimit");
-  pcre_exc_Backtrack     = caml_named_value("Pcre.Backtrack");
+  pcre_exc_Error     = caml_named_value("Pcre.Error");
+  pcre_exc_Backtrack = caml_named_value("Pcre.Backtrack");
 
-  var_Start_only         = hash_variant("Start_only");
-  var_ANCHORED           = hash_variant("ANCHORED");
-  var_Char               = hash_variant("Char");
-  var_Not_studied        = hash_variant("Not_studied");
-  var_Studied            = hash_variant("Studied");
-  var_Optimal            = hash_variant("Optimal");
+  var_Start_only  = caml_hash_variant("Start_only");
+  var_ANCHORED    = caml_hash_variant("ANCHORED");
+  var_Char        = caml_hash_variant("Char");
+  var_Not_studied = caml_hash_variant("Not_studied");
+  var_Studied     = caml_hash_variant("Studied");
+  var_Optimal     = caml_hash_variant("Optimal");
 
   pcre_callout = &pcre_callout_handler;
 
@@ -155,28 +165,62 @@ static void pcre_dealloc_regexp(value v_rex)
   if (extra != NULL) (pcre_free)(extra);
 }
 
-/* Raises exceptions which take two arguments */
-static void raise_with_two_args(value tag, value arg1, value arg2)
-{
-  value v_exc;
-
-  /* Protects tag, arg1 and arg2 from being reclaimed by the garbage
-     collector when the exception value is allocated */
-  Begin_roots3(tag, arg1, arg2);
-    v_exc = caml_alloc_small(3, 0);
-    Field(v_exc, 0) = tag;
-    Field(v_exc, 1) = arg1;
-    Field(v_exc, 2) = arg2;
-  End_roots();
-
-  caml_raise(v_exc);
-}
-
 /* Makes OCaml-string from PCRE-version */
-CAMLprim value pcre_version_stub(value unit) {
+CAMLprim value pcre_version_stub(value __unused v_unit)
+{
   return caml_copy_string((char *) pcre_version());
 }
 
+
+/* Raising exceptions */
+
+static inline void raise_pcre_error(value v_arg) Noreturn;
+static inline void raise_partial() Noreturn;
+static inline void raise_bad_partial() Noreturn;
+static inline void raise_bad_utf8() Noreturn;
+static inline void raise_bad_utf8_offset() Noreturn;
+static inline void raise_match_limit() Noreturn;
+static inline void raise_recursion_limit() Noreturn;
+static inline void raise_bad_pattern(const char *msg, int pos) Noreturn;
+static inline void raise_internal_error(char *msg) Noreturn;
+
+static inline void raise_pcre_error(value v_arg)
+{ caml_raise_with_arg(*pcre_exc_Error, v_arg); }
+
+static inline void raise_partial() { raise_pcre_error(Val_int(0)); }
+static inline void raise_bad_partial() { raise_pcre_error(Val_int(1)); }
+static inline void raise_bad_utf8() { raise_pcre_error(Val_int(2)); }
+static inline void raise_bad_utf8_offset() { raise_pcre_error(Val_int(3)); }
+static inline void raise_match_limit() { raise_pcre_error(Val_int(4)); }
+static inline void raise_recursion_limit() { raise_pcre_error(Val_int(5)); }
+
+static inline void raise_bad_pattern(const char *msg, int pos)
+{
+  CAMLparam0();
+  CAMLlocal1(v_msg);
+  value v_arg;
+  v_msg = caml_copy_string(msg);
+  v_arg = caml_alloc_small(2, 0);
+  Field(v_arg, 0) = v_msg;
+  Field(v_arg, 1) = Val_int(pos);
+  raise_pcre_error(v_arg);
+  CAMLnoreturn;
+}
+
+static inline void raise_internal_error(char *msg)
+{
+  CAMLparam0();
+  CAMLlocal1(v_msg);
+  value v_arg;
+  v_msg = caml_copy_string(msg);
+  v_arg = caml_alloc_small(1, 1);
+  Field(v_arg, 0) = v_msg;
+  raise_pcre_error(v_arg);
+  CAMLnoreturn;
+}
+
+/* PCRE pattern compilation */
+
 /* Makes compiled regular expression from compilation options, an optional
    value of chartables and the pattern string */
 CAMLprim value pcre_compile_stub(value v_opt, value v_tables, value v_pat)
@@ -194,11 +238,9 @@ CAMLprim value pcre_compile_stub(value v_opt, value v_tables, value v_pat)
   pcre *regexp = pcre_compile(String_val(v_pat), Int_val(v_opt), &error,
                               &error_ofs, tables);
 
-  /* Raises appropriate exception [BadPattern] if the pattern could not
-     be compiled */
-  if (regexp == NULL) raise_with_two_args(*pcre_exc_BadPattern,
-                                          caml_copy_string((char *) error),
-                                          Val_int(error_ofs));
+  /* Raises appropriate exception with [BadPattern] if the pattern
+     could not be compiled */
+  if (regexp == NULL) raise_bad_pattern(error, error_ofs);
 
   /* Finalized value: GC will do a full cycle every 500 regexp allocations
      (one regexp consumes in average probably less than 100 bytes ->
@@ -227,22 +269,52 @@ CAMLprim value pcre_study_stub(value v_rex)
   if (! (int) Field(v_rex, 3)) {
     const char *error = NULL;
     pcre_extra *extra = pcre_study((pcre *) Field(v_rex, 1), 0, &error);
-    if (error != NULL) invalid_argument((char *) error);
+    if (error != NULL) caml_invalid_argument((char *) error);
     Field(v_rex, 2) = (value) extra;
     Field(v_rex, 3) = Val_int(1);
   }
   return v_rex;
 }
 
+/* Sets a match limit recursion for a regular expression imperatively */
+CAMLprim value pcre_set_imp_match_limit_recursion_stub(value v_rex, value v_lim)
+{
+  pcre_extra *extra = (pcre_extra *) Field(v_rex, 2);
+  if (extra == NULL) {
+    extra = caml_stat_alloc(sizeof(pcre_extra));
+    extra->flags = PCRE_EXTRA_MATCH_LIMIT_RECURSION;
+    Field(v_rex, 2) = (value) extra;
+  } else {
+    unsigned long int *flags_ptr = &extra->flags;
+    *flags_ptr = PCRE_EXTRA_MATCH_LIMIT_RECURSION | *flags_ptr;
+  }
+  extra->match_limit_recursion = Int_val(v_lim);
+  return v_rex;
+}
+
+/* Gets the match limit recursion of a regular expression if it exists */
+CAMLprim value pcre_get_match_limit_recursion_stub(value v_rex)
+{
+  pcre_extra *extra = (pcre_extra *) Field(v_rex, 2);
+  if (extra == NULL) return None;
+  if (extra->flags & PCRE_EXTRA_MATCH_LIMIT_RECURSION) {
+    value v_lim = Val_int(extra->match_limit_recursion);
+    value v_res = caml_alloc_small(1, 0);
+    Field(v_res, 0) = v_lim;
+    return v_res;
+  }
+  return None;
+}
+
 /* Sets a match limit for a regular expression imperatively */
-CAMLprim value pcre_set_imp_match_limit_stub(value v_rex, value v_lim){
+CAMLprim value pcre_set_imp_match_limit_stub(value v_rex, value v_lim)
+{
   pcre_extra *extra = (pcre_extra *) Field(v_rex, 2);
   if (extra == NULL) {
     extra = caml_stat_alloc(sizeof(pcre_extra));
     extra->flags = PCRE_EXTRA_MATCH_LIMIT;
     Field(v_rex, 2) = (value) extra;
-  }
-  else {
+  } else {
     unsigned long int *flags_ptr = &extra->flags;
     *flags_ptr = PCRE_EXTRA_MATCH_LIMIT | *flags_ptr;
   }
@@ -251,20 +323,21 @@ CAMLprim value pcre_set_imp_match_limit_stub(value v_rex, value v_lim){
 }
 
 /* Gets the match limit of a regular expression if it exists */
-CAMLprim value pcre_get_match_limit_stub(value v_rex){
+CAMLprim value pcre_get_match_limit_stub(value v_rex)
+{
   pcre_extra *extra = (pcre_extra *) Field(v_rex, 2);
   if (extra == NULL) return None;
   if (extra->flags & PCRE_EXTRA_MATCH_LIMIT) {
-    value lim = Val_int(extra->match_limit);
-    value res = caml_alloc_small(1, 0);
-    Field(res, 0) = lim;
-    return res;
+    value v_lim = Val_int(extra->match_limit);
+    value v_res = caml_alloc_small(1, 0);
+    Field(v_res, 0) = v_lim;
+    return v_res;
   }
   return None;
 }
 
 /* Performs the call to the pcre_fullinfo function */
-static value pcre_fullinfo_stub(value v_rex, int what, void *where)
+static inline int pcre_fullinfo_stub(value v_rex, int what, void *where)
 {
   return pcre_fullinfo((pcre *) Field(v_rex, 1), (pcre_extra *) Field(v_rex, 2),
                        what, where);
@@ -273,38 +346,36 @@ static value pcre_fullinfo_stub(value v_rex, int what, void *where)
 /* Some stubs for info-functions */
 
 /* Generic macro for getting integer results from pcre_fullinfo */
-#define make_int_info(name, option) \
+#define make_info(tp, cnv, name, option) \
   CAMLprim value pcre_##name##_stub(value v_rex) \
   { \
-    int options; \
+    tp options; \
     const int ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_##option, &options); \
-    if (ret != 0) \
-      raise_with_string(*pcre_exc_InternalError, "pcre_##name##_stub"); \
-    return Val_int(options); \
+    if (ret != 0) raise_internal_error("pcre_##name##_stub"); \
+    return cnv(options); \
   }
 
-make_int_info(options, OPTIONS)
-make_int_info(size, SIZE)
-make_int_info(studysize, STUDYSIZE)
-make_int_info(capturecount, CAPTURECOUNT)
-make_int_info(backrefmax, BACKREFMAX)
-make_int_info(namecount, NAMECOUNT)
-make_int_info(nameentrysize, NAMEENTRYSIZE)
+make_info(unsigned long int, Val_long, options, OPTIONS)
+make_info(size_t, Val_long, size, SIZE)
+make_info(size_t, Val_long, studysize, STUDYSIZE)
+make_info(int, Val_int, capturecount, CAPTURECOUNT)
+make_info(int, Val_int, backrefmax, BACKREFMAX)
+make_info(int, Val_int, namecount, NAMECOUNT)
+make_info(int, Val_int, nameentrysize, NAMEENTRYSIZE)
 
 CAMLprim value pcre_firstbyte_stub(value v_rex)
 {
   int firstbyte;
   const int ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_FIRSTBYTE, &firstbyte);
 
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError,
-                                  "pcre_firstbyte_stub");
+  if (ret != 0) raise_internal_error("pcre_firstbyte_stub");
 
   switch (firstbyte) {
     case -1 : return var_Start_only; break;  /* [`Start_only] */
     case -2 : return var_ANCHORED; break;    /* [`ANCHORED] */
     default :
       if (firstbyte < 0 )  /* Should not happen */
-        raise_with_string(*pcre_exc_InternalError, "pcre_firstbyte_stub");
+        raise_internal_error("pcre_firstbyte_stub");
       else {
         value v_firstbyte;
         /* Allocates the non-constant constructor [`Char of char] and fills
@@ -324,8 +395,7 @@ CAMLprim value pcre_firsttable_stub(value v_rex)
   int ret =
     pcre_fullinfo_stub(v_rex, PCRE_INFO_FIRSTTABLE, (void *) &ftable);
 
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError,
-                                  "pcre_firsttable_stub");
+  if (ret != 0) raise_internal_error("pcre_firsttable_stub");
 
   if (ftable == NULL) return None;
   else {
@@ -357,12 +427,10 @@ CAMLprim value pcre_lastliteral_stub(value v_rex)
   const int ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_LASTLITERAL,
                                         &lastliteral);
 
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError,
-                                  "pcre_lastliteral_stub");
+  if (ret != 0) raise_internal_error("pcre_lastliteral_stub");
 
   if (lastliteral == -1) return None;
-  if (lastliteral < 0) raise_with_string(*pcre_exc_InternalError,
-                                         "pcre_lastliteral_stub");
+  if (lastliteral < 0) raise_internal_error("pcre_lastliteral_stub");
   else {
     /* Allocates [Some char] */
     value v_res = caml_alloc_small(1, 0);
@@ -381,6 +449,28 @@ CAMLprim value pcre_study_stat_stub(value v_rex)
   return var_Not_studied;  /* otherwise [`Not_studied] */
 }
 
+static inline void handle_exec_error(char *loc, const int ret) Noreturn;
+
+static inline void handle_exec_error(char *loc, const int ret)
+{
+  switch (ret) {
+    /* Dedicated exceptions */
+    case PCRE_ERROR_NOMATCH : caml_raise_not_found();
+    case PCRE_ERROR_PARTIAL : raise_partial();
+    case PCRE_ERROR_MATCHLIMIT : raise_match_limit();
+    case PCRE_ERROR_BADPARTIAL : raise_bad_partial();
+    case PCRE_ERROR_BADUTF8 : raise_bad_utf8();
+    case PCRE_ERROR_BADUTF8_OFFSET : raise_bad_utf8_offset();
+    case PCRE_ERROR_RECURSIONLIMIT : raise_recursion_limit();
+    /* Unknown error */
+    default : {
+      char err_buf[100];
+      snprintf(err_buf, 100, "%s: unhandled PCRE error code: %d", loc, ret);
+      raise_internal_error(err_buf);
+    }
+  }
+}
+
 /* Executes a pattern match with runtime options, a regular expression, a
    string offset, a string length, a subject string, a number of subgroup
    offsets, an offset vector and an optional callout function */
@@ -388,10 +478,10 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
                               value v_subj, value v_subgroups2, value v_ovec,
                               value v_maybe_cof)
 {
-  const int ofs = Int_val(v_ofs), len = string_length(v_subj);
+  const int ofs = Int_val(v_ofs), len = caml_string_length(v_subj);
 
   if (ofs > len || ofs < 0)
-    invalid_argument("Pcre.pcre_exec_stub: illegal offset");
+    caml_invalid_argument("Pcre.pcre_exec_stub: illegal offset");
 
   {
     const pcre *code = (pcre *) Field(v_rex, 1);  /* Compiled pattern */
@@ -410,23 +500,14 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
       const int ret =
         pcre_exec(code, extra, ocaml_subj, len, ofs, opt, ovec, subgroups3);
 
-      if (ret < 0) {
-        switch(ret) {
-          case PCRE_ERROR_NOMATCH : raise_constant(*pcre_exc_Not_found);
-          case PCRE_ERROR_PARTIAL : raise_constant(*pcre_exc_Partial);
-          case PCRE_ERROR_MATCHLIMIT : raise_constant(*pcre_exc_MatchLimit);
-          case PCRE_ERROR_BADPARTIAL : raise_constant(*pcre_exc_BadPartial);
-          case PCRE_ERROR_BADUTF8 : raise_constant(*pcre_exc_BadUTF8);
-          case PCRE_ERROR_BADUTF8_OFFSET :
-            raise_constant(*pcre_exc_BadUTF8Offset);
-          default :
-            raise_with_string(*pcre_exc_InternalError, "pcre_exec_stub");
-        }
-      }
-
+      if (ret < 0) handle_exec_error("pcre_exec_stub", ret);
       else {
         const int *ovec_src = ovec + subgroups2_1;
+#if _WIN64
+        long long *ovec_dst = (long long *) ovec + subgroups2_1;
+#else
         long int *ovec_dst = (long int *) ovec + subgroups2_1;
+#endif
 
         /* Converts offsets from C-integers to OCaml-Integers
            This is a bit tricky, because there are 32- and 64-bit platforms
@@ -447,7 +528,20 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
       int *ovec = caml_stat_alloc(sizeof(int) * subgroups3);
       int ret;
       struct cod cod = { (value *) NULL, (value *) NULL, (value) NULL };
-      struct pcre_extra new_extra = { PCRE_EXTRA_CALLOUT_DATA, NULL, 0, NULL };
+      struct pcre_extra new_extra =
+#ifdef PCRE_EXTRA_MATCH_LIMIT_RECURSION
+# ifdef PCRE_EXTRA_MARK
+#  ifdef PCRE_EXTRA_EXECUTABLE_JIT
+        { PCRE_EXTRA_CALLOUT_DATA, NULL, 0, NULL, NULL, 0, NULL, NULL };
+#  else
+        { PCRE_EXTRA_CALLOUT_DATA, NULL, 0, NULL, NULL, 0, NULL };
+#  endif
+# else
+        { PCRE_EXTRA_CALLOUT_DATA, NULL, 0, NULL, NULL, 0 };
+# endif
+#else
+        { PCRE_EXTRA_CALLOUT_DATA, NULL, 0, NULL, NULL };
+#endif
 
       memcpy(subj, ocaml_subj, len);
 
@@ -471,31 +565,22 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
           new_extra.flags = PCRE_EXTRA_CALLOUT_DATA | extra->flags;
           new_extra.study_data = extra->study_data;
           new_extra.match_limit = extra->match_limit;
+          new_extra.tables = extra->tables;
+#ifdef PCRE_EXTRA_MATCH_LIMIT_RECURSION
+          new_extra.match_limit_recursion = extra->match_limit_recursion;
+#endif
 
           ret = pcre_exec(code, &new_extra, subj, len, ofs, opt, ovec,
                           subgroups3);
         }
 
-        free(subj);
+        caml_stat_free(subj);
       End_roots();
 
       if (ret < 0) {
-        free(ovec);
-        switch(ret) {
-          case PCRE_ERROR_NOMATCH : raise_constant(*pcre_exc_Not_found);
-          case PCRE_ERROR_PARTIAL : raise_constant(*pcre_exc_Partial);
-          case PCRE_ERROR_MATCHLIMIT : raise_constant(*pcre_exc_MatchLimit);
-          case PCRE_ERROR_BADPARTIAL : raise_constant(*pcre_exc_BadPartial);
-          case PCRE_ERROR_BADUTF8 : raise_constant(*pcre_exc_BadUTF8);
-          case PCRE_ERROR_BADUTF8_OFFSET :
-            raise_constant(*pcre_exc_BadUTF8Offset);
-          case PCRE_ERROR_CALLOUT : caml_raise(cod.v_exn);
-          default :
-            raise_with_string(*pcre_exc_InternalError, "pcre_exec_stub");
-        }
-      }
-
-      else {
+        caml_stat_free(ovec);
+        handle_exec_error("pcre_exec_stub(callout)", ret);
+      } else {
         int *ovec_src = ovec + subgroups2_1;
         long int *ovec_dst = &Field(v_ovec, 0) + subgroups2_1;
 
@@ -504,7 +589,7 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
           --ovec_src; --ovec_dst;
         }
 
-        free(ovec);
+        caml_stat_free(ovec);
       }
     }
   }
@@ -514,7 +599,7 @@ CAMLprim value pcre_exec_stub(value v_opt, value v_rex, value v_ofs,
 
 /* Byte-code hook for pcre_exec_stub
    Needed, because there are more than 5 arguments */
-CAMLprim value pcre_exec_stub_bc(value *argv, int argn)
+CAMLprim value pcre_exec_stub_bc(value *argv, int __unused argn)
 {
   return pcre_exec_stub(argv[0], argv[1], argv[2], argv[3],
                         argv[4], argv[5], argv[6]);
@@ -522,7 +607,7 @@ CAMLprim value pcre_exec_stub_bc(value *argv, int argn)
 
 /* Generates a new set of chartables for the current locale (see man
    page of PCRE */
-CAMLprim value pcre_maketables_stub(value unit)
+CAMLprim value pcre_maketables_stub(value __unused v_unit)
 {
   /* GC will do a full cycle every 100 table set allocations
      (one table set consumes 864 bytes -> maximum of 86400 bytes
@@ -543,7 +628,9 @@ CAMLprim value pcre_get_stringnumber_stub(value v_rex, value v_name)
 {
   const int ret = pcre_get_stringnumber((pcre *) Field(v_rex, 1),
                                         String_val(v_name));
-  if (ret == PCRE_ERROR_NOSUBSTRING) invalid_argument("Named string not found");
+  if (ret == PCRE_ERROR_NOSUBSTRING)
+    caml_invalid_argument("Named string not found");
+
   return Val_int(ret);
 }
 
@@ -558,13 +645,13 @@ CAMLprim value pcre_names_stub(value v_rex)
   int i;
 
   int ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_NAMECOUNT, &name_count);
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError, "pcre_names_stub");
+  if (ret != 0) raise_internal_error("pcre_names_stub: namecount");
 
   ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_NAMEENTRYSIZE, &entry_size);
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError, "pcre_names_stub");
+  if (ret != 0) raise_internal_error("pcre_names_stub: nameentrysize");
 
   ret = pcre_fullinfo_stub(v_rex, PCRE_INFO_NAMETABLE, &tbl_ptr);
-  if (ret != 0) raise_with_string(*pcre_exc_InternalError, "pcre_names_stub");
+  if (ret != 0) raise_internal_error("pcre_names_stub: nametable");
 
   v_res = caml_alloc(name_count, 0);
 
@@ -578,31 +665,43 @@ CAMLprim value pcre_names_stub(value v_rex)
 }
 
 /* Generic stub for getting integer results from pcre_config */
-static int pcre_config_int(int what)
+static inline int pcre_config_int(int what)
 {
   int ret;
   pcre_config(what, (void *) &ret);
   return ret;
 }
 
+/* Generic stub for getting long integer results from pcre_config */
+static inline int pcre_config_long(int what)
+{
+  long ret;
+  pcre_config(what, (void *) &ret);
+  return ret;
+}
+
 /* Some stubs for config-functions */
 
 /* Returns boolean indicating UTF8-support */
-CAMLprim value pcre_config_utf8_stub(value unit)
+CAMLprim value pcre_config_utf8_stub(value __unused v_unit)
 { return Val_bool(pcre_config_int(PCRE_CONFIG_UTF8)); }
 
 /* Returns character used as newline */
-CAMLprim value pcre_config_newline_stub(value unit)
+CAMLprim value pcre_config_newline_stub(value __unused v_unit)
 { return Val_int(pcre_config_int(PCRE_CONFIG_NEWLINE)); }
 
 /* Returns number of bytes used for internal linkage of regular expressions */
-CAMLprim value pcre_config_link_size_stub(value unit)
+CAMLprim value pcre_config_link_size_stub(value __unused v_unit)
 { return Val_int(pcre_config_int(PCRE_CONFIG_LINK_SIZE)); }
 
-/* Returns default limit for calls to internal matching function */
-CAMLprim value pcre_config_match_limit_stub(value unit)
-{ return Val_int(pcre_config_int(PCRE_CONFIG_MATCH_LIMIT)); }
-
 /* Returns boolean indicating use of stack recursion */
-CAMLprim value pcre_config_stackrecurse_stub(value unit)
+CAMLprim value pcre_config_stackrecurse_stub(value __unused v_unit)
 { return Val_bool(pcre_config_int(PCRE_CONFIG_STACKRECURSE)); }
+
+/* Returns default limit for calls to internal matching function */
+CAMLprim value pcre_config_match_limit_stub(value __unused v_unit)
+{ return Val_long(pcre_config_long(PCRE_CONFIG_MATCH_LIMIT)); }
+
+/* Returns default limit for calls to internal matching function */
+CAMLprim value pcre_config_match_limit_recursion_stub(value __unused v_unit)
+{ return Val_long(pcre_config_long(PCRE_CONFIG_MATCH_LIMIT_RECURSION)); }
